@@ -43,7 +43,7 @@
 #endif
 #include <sys/socket.h>
 
-#if defined(__FreeBSD__) || defined(__DragonFly__)
+#if defined(__FreeBSD__) || defined(__DragonFly__) || defined(__APPLE__)
 #include <signal.h>
 #include <sys/resource.h>
 #ifndef RUSAGE_SELF
@@ -109,7 +109,8 @@ error_client( MsgChannel *client, string error )
 }
 
 int work_it( CompileJob &j, unsigned int job_stat[], MsgChannel* client,
-             CompileResultMsg& rmsg, string &outfilename, unsigned long int mem_limit, int client_fd )
+             CompileResultMsg& rmsg, const string &outfilename, 
+             unsigned long int mem_limit, int client_fd, int job_in_fd )
 {
     rmsg.out.erase(rmsg.out.begin(), rmsg.out.end());
     rmsg.out.erase(rmsg.out.begin(), rmsg.out.end());
@@ -118,14 +119,6 @@ int work_it( CompileJob &j, unsigned int job_stat[], MsgChannel* client,
     appendList( list, j.restFlags() );
     int ret;
 
-    char tmp_output[PATH_MAX];
-    char prefix_output[PATH_MAX]; // I'm too lazy to calculate how many digits 2^64 is :)
-    sprintf( prefix_output, "icecc-%d", j.jobID() );
-    if ( ( ret = dcc_make_tmpnam(prefix_output, ".o", tmp_output, 1 ) ) != 0 )
-        return ret;
-
-    outfilename = tmp_output;
-
     int sock_err[2];
     int sock_out[2];
     int sock_in[2];
@@ -196,7 +189,6 @@ int work_it( CompileJob &j, unsigned int job_stat[], MsgChannel* client,
         close( sock_in[0] );
         close( sock_in[1] );
         close( sock_err[0] );
-        unlink( tmp_output );
         return EXIT_OUT_OF_MEMORY;
     } else if ( pid == 0 ) {
 
@@ -257,7 +249,7 @@ int work_it( CompileJob &j, unsigned int job_stat[], MsgChannel* client,
         argv[i++] = strdup((j.language() == CompileJob::Lang_CXX) ? "c++" : "c");
         argv[i++] = strdup( "-" );
         argv[i++] = strdup( "-o" );
-        argv[i++] = tmp_output;
+        argv[i++] = strdup(outfilename.c_str());
         argv[i++] = strdup( "--param" );
         sprintf( buffer, "ggc-min-expand=%d", ggc_min_expand_heuristic( mem_limit ) );
         argv[i++] = strdup( buffer );
@@ -327,6 +319,8 @@ int work_it( CompileJob &j, unsigned int job_stat[], MsgChannel* client,
                 if ( bytes < 0 && errno == EINTR )
                     continue;
 
+                write(job_in_fd, fcmsg->buffer + off, bytes);
+
                 if ( bytes == -1 ) {
                     log_perror("write to caching socket failed. ");
 
